/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TEncAdaptiveFilter.cpp
 \brief    estimation part of adaptive filter class
 */
#include "TEncAdaptiveFilter.h"

#if SVC_IL_ADAPTIVE_FILTER

// #include <string.h>
// #include <stdlib.h>
// #include <stdio.h>
#include <math.h>

//! \ingroup TLibEncoder
//! \{

// ====================================================================================================================
// Constructor / destructor / create / destroy / operator
// ====================================================================================================================

TEncAdaptiveFilter::TEncAdaptiveFilter()
{
  m_ctuEnableFlag = NULL;
  m_ctuInfo = NULL;
}

TEncAdaptiveFilter::~TEncAdaptiveFilter()
{
  destroyAFEncData();
}

Void TEncAdaptiveFilter::createAFEncData() 
{
  Int numCTUs = m_numCTUInHeight*m_numCTUInWidth;
  m_ctuEnableFlag = new Bool[numCTUs];
  //default enable all CTUs
  for(Int i=0; i < numCTUs; i++)
  {
    m_ctuEnableFlag[i] = true;
  }
  m_ctuInfo       = new AFCTUInfo[numCTUs];
}

Void TEncAdaptiveFilter::destroyAFEncData() 
{
  if(m_ctuEnableFlag)
  {
    delete[] m_ctuEnableFlag;
    m_ctuEnableFlag = NULL;
  }

  if(m_ctuInfo)
  {
    delete[] m_ctuInfo;
    m_ctuInfo = NULL;
  }

}


// ====================================================================================================================
// methods
// ====================================================================================================================

#if SVC_ILAF_QD_MODEL
Void TEncAdaptiveFilter::estimateEarlyTermination(TComPicYuv* orgYuv, TComPicYuv* upYuv, Int baseQp, SliceType sliceType, Int enhanceQp,  Bool &earlyTermination)
{
  //Double dQPDis = 0.225*(baseQp-12)*(baseQp-12);
  Double dQPDis = 0.3*(baseQp-12)*(baseQp-12);

  Int orgStride = orgYuv->getStride();
  Int upStride  = upYuv->getStride();
  Int width = orgYuv->getWidth();
  Int height = orgYuv->getHeight();
  Pel *orgLuma = orgYuv->getLumaAddr();
  Pel *upLuma  = upYuv->getLumaAddr();
  Int iSum_Distortion = 0;

  for (Int i=0; i<height; i++)
  {
    for (Int j=0; j<width ;j++)
    {
      iSum_Distortion += (orgLuma[j]-upLuma[j]) * (orgLuma[j]-upLuma[j]);
    }
    orgLuma += orgStride;
    upLuma += upStride;
  }

  iSum_Distortion /= ((width*height)>>2);
  //  if ( (sliceType !=I_SLICE && iSum_Distortion>dQPDis) || (enhanceQp >= baseQp+4 && iSum_Distortion<dQPDis) )
  if( ( enhanceQp >= baseQp+4 ) || ( iSum_Distortion > dQPDis)   )
  {
    earlyTermination = true;
  }
  else 
  {
    earlyTermination = false;
  }
}
#endif

#if SVC_ILAF_ONOFF_FLAG
Void TEncAdaptiveFilter::filterProcess(TComPicYuv* orgYuv, TComPicYuv* upYuv, TComPicYuv* filteredYuv, Double lambdaLuma, Double lambdaChroma)
#else
Void TEncAdaptiveFilter::filterProcess(TComPicYuv* orgYuv, TComPicYuv* upYuv, Double lambdaLuma, Double lambdaChroma)
#endif
{
  m_labmda[AF_Y]= lambdaLuma; m_labmda[AF_Cb]= lambdaChroma; m_labmda[AF_Cr]= lambdaChroma;

  TComPicYuv* srcYuv = m_tempPicYuv;
  upYuv->copyToPic(srcYuv);
  srcYuv->setBorderExtension(false);
  srcYuv->extendPicBorder();

  //prepare encode data
  prepareEncData();

  //gather correlations for all CTB
  getStatistics(m_afCorr, orgYuv, srcYuv);

  //estimate filters for each CTB
  estimateFilters(m_afCorr, m_afFilter);

  //filtering
#if SVC_ILAF_ONOFF_FLAG
  filterPic(srcYuv, filteredYuv, m_afFilter);
#else
  filterPic(srcYuv, upYuv, m_afFilter);
#endif
}

Void TEncAdaptiveFilter::prepareEncData()
{
  //prepare LCU information
  for(Int j=0; j< m_numCTUInHeight; j++)
  {
    for(Int i=0; i< m_numCTUInWidth; i++)
    {
      Int addr = j*m_numCTUInWidth + i;
      AFCTUInfo* ctuInfo= &m_ctuInfo[addr];

      ctuInfo->yPos  = j* m_maxCUHeight;
      ctuInfo->xPos  = i* m_maxCUWidth;
      ctuInfo->height= (ctuInfo->yPos + m_maxCUHeight > m_picHeight)?( m_picHeight- ctuInfo->yPos ):m_maxCUHeight;
      ctuInfo->width = (ctuInfo->xPos + m_maxCUWidth  > m_picWidth )?( m_picWidth - ctuInfo->xPos ):m_maxCUWidth ;
    }
  }
}

Void TEncAdaptiveFilter::getStatistics(AFCorrData* afCorr, TComPicYuv* orgYuv, TComPicYuv* srcYuv)
{
  Int numCTUs= m_numCTUInHeight*m_numCTUInWidth;

  //find last valid block
  Int lastBlockIdx =-1;
  for(Int addr=0; addr< numCTUs; addr++)
  {
    if(m_ctuEnableFlag[addr])
    {
      lastBlockIdx= addr;
    }
  }

  //gather correlation
  for(Int compIdx=0; compIdx < AF_NUM_COMPONETS; compIdx++)
  {
    Int         formatShift= (compIdx == AF_Y)?0:1;
    Pel*        orgPic = getPicBuf(orgYuv, compIdx);
    Pel*        srcPic = getPicBuf(srcYuv, compIdx);
    Int         stride = (compIdx == AF_Y)?orgYuv->getStride():orgYuv->getCStride();
    AFCorrData* corr = &afCorr[compIdx];

    corr->reset();

    for(Int addr=0; addr< numCTUs; addr++)
    {
      AFCTUInfo* ctuInfo= &m_ctuInfo[addr];
      if(m_ctuEnableFlag[addr] || lastBlockIdx==-1 )
      {
        getBlockCorr(corr, orgPic, srcPic, stride
                    , (ctuInfo->yPos >> formatShift)
                    , (ctuInfo->xPos >> formatShift)
                    , (ctuInfo->height >> formatShift)
                    , (ctuInfo->width >> formatShift)
                    , (lastBlockIdx == -1)?( addr == numCTUs-1 ):( addr== lastBlockIdx  )
                    );
      }
    }
  }
}

Void TEncAdaptiveFilter::estimateFilters(AFCorrData* afCorr, AFFilter* afFilter)
{
  for(Int compIdx=0; compIdx< AF_NUM_COMPONETS; compIdx++)
  {
    AFCorrData* corr= &afCorr[compIdx];
    Double      lambda = m_labmda[compIdx];

    //RD cost (enabled)
    AFFilter& encFilt= afFilter[compIdx];
    encFilt.enableFlag= true;
    encFilt.filterShape= 0; //default
    deriveFilter(corr, encFilt.coeff);
    Int64 encDist= estimateDistReduction(corr, encFilt.coeff);
    predictCoef(encFilt.coeff);
    Int encRate= estimateFiltRate(&encFilt);
#if SVC_ILAF_ONOFF_FLAG
    Double encCost = (Double)encDist + lambda*((Double)encRate+(Double)(m_numCTUInHeight*m_numCTUInWidth)*(Double)0.6);
#else
    Double encCost = (Double)encDist + lambda*((Double)encRate);
#endif

    //RD cost (disabled)
    AFFilter offFilt;
    offFilt.enableFlag = false;
    offFilt.filterShape = 0; //default
    deriveOffFilter(offFilt.coeff);
    Int64 offDist= estimateDistReduction(corr, offFilt.coeff);
    Int offRate= estimateFiltRate(&offFilt);
    Double offCost = (Double)offDist + lambda*((Double)offRate);

    if(offCost <= encCost)
    {
      encFilt = offFilt;
    }
  }
}

Void TEncAdaptiveFilter::filterPic(TComPicYuv* srcYuv, TComPicYuv* resYuv, AFFilter* afFilter)
{
  Int coef[AF_MAX_NUM_COEF];
  for(Int compIdx=0; compIdx< AF_NUM_COMPONETS; compIdx++)
  {
    Pel*  srcPic = getPicBuf(srcYuv, compIdx);
    Pel*  resPic = getPicBuf(resYuv, compIdx);
    Int   stride = (compIdx == AF_Y)? srcYuv->getStride():srcYuv->getCStride();
    Int   formatShift= (compIdx == AF_Y)?0:1;

    if(afFilter[compIdx].enableFlag)
    {
      reconstructCoef(coef, afFilter[compIdx].coeff);
      filterBlock(afFilter[compIdx].filterShape, coef, srcPic, resPic, stride, stride, (m_picWidth >> formatShift), (m_picHeight >> formatShift));
    }
    else
    {
      if (compIdx == AF_Y)
      {
        srcYuv->copyToPicLuma(resYuv);
      }
      else if (compIdx == AF_Cb)
      {
        srcYuv->copyToPicCb(resYuv);
      }
      else
      {
        srcYuv->copyToPicCr(resYuv);
      }      
    }
  }
}


Void TEncAdaptiveFilter::getBlockCorr(AFCorrData* afCorr
                                     , Pel* orgPic, Pel* srcPic, Int stride
                                     , Int yPos, Int xPos, Int height, Int width
                                     , Bool isLastBlock
                                     , Int shape)
{

  Pel* imgPad = srcPic + (yPos*stride);
  Pel* imgOrg = orgPic + (yPos*stride);
  Int yPosEnd = yPos + height;
  Int xPosEnd = xPos + width;
  Int N = g_AFNumCoef[shape];
  Double** eCorr = afCorr->ECorr;
  Double*  yCorr = afCorr->yCorr;
  Double&  pixAcc= afCorr->pixAcc;

  Int ELocal[AF_MAX_NUM_COEF];
  Int i, j, k, l, yLocal;

  switch(shape)
  {
  case AF_SHAPE_7x7CROSS_3x3SQUARE:
    {
      Pel *imgPad1, *imgPad2, *imgPad3, *imgPad4, *imgPad5, *imgPad6;
      for (i= yPos; i< yPosEnd; i++)
      {
        imgPad1 = imgPad + stride;
        imgPad2 = imgPad - stride;
        imgPad3 = imgPad + 2*stride;
        imgPad4 = imgPad - 2*stride;
        imgPad5 = imgPad + 3*stride;
        imgPad6 = imgPad - 3*stride;

        for (j= xPos; j< xPosEnd; j++)
        {
          memset(ELocal, 0, N*sizeof(Int));

          ELocal[0] = (imgPad5[j] + imgPad6[j]);

          ELocal[1] = (imgPad3[j] + imgPad4[j]);

          ELocal[2] = (imgPad1[j+1] + imgPad2[j-1]);
          ELocal[3] = (imgPad1[j  ] + imgPad2[j  ]);
          ELocal[4] = (imgPad1[j-1] + imgPad2[j+1]);

          ELocal[5] = (imgPad[j+3] + imgPad[j-3]);
          ELocal[6] = (imgPad[j+2] + imgPad[j-2]);
          ELocal[7] = (imgPad[j+1] + imgPad[j-1]);
          ELocal[8] = (imgPad[j  ]);

          yLocal= (Int)imgOrg[j];
          pixAcc += yLocal;

          for(k=0; k<N; k++)
          {
            eCorr[k][k] += ELocal[k]*ELocal[k];
            for(l=k+1; l<N; l++)
            {
              eCorr[k][l] += ELocal[k]*ELocal[l];
            }

            yCorr[k] += yLocal*ELocal[k];
          }

        }
        imgPad+= stride;
        imgOrg+= stride;

      }
    }
    break;
  default:
    {
      printf("Not a supported filter shape\n");
      assert(0);
      exit(-1);
    }

  }

  if(isLastBlock)
  {
    for(j=0; j<N-1; j++)
    {
      for(i=j+1; i<N; i++)
      {
        eCorr[i][j] = eCorr[j][i];
      }
    }
  }


}

Void TEncAdaptiveFilter::deriveFilter(AFCorrData* afCorr, Int* coef, Int shape)
{
  Double fracCoeff[AF_MAX_NUM_COEF];
  Int  numCoef = g_AFNumCoef[shape];
  Int* filtMag = m_shapeWeight[shape]; 

  gnsSolveByChol(afCorr->ECorr, afCorr->yCorr, fracCoeff, numCoef);
  quantFilterCoef(fracCoeff, coef, filtMag, numCoef);
}

Void TEncAdaptiveFilter::deriveOffFilter(Int* coef, Int shape)
{
  Int  numCoef = g_AFNumCoef[shape];
  ::memset(coef, 0, sizeof(Int)*numCoef);
  coef[numCoef-1]= (1 << AF_NUM_BIT_SHIFT);
}

Int64 TEncAdaptiveFilter::estimateDistReduction(AFCorrData* afCorr, Int* coef, Int shape )
{

  Int  numCoef = g_AFNumCoef[shape];
  Double**  eCorr   = afCorr->ECorr;
  Double*   yCorr   = afCorr->yCorr;

  
  Double dcoeff[AF_MAX_NUM_COEF];

  //variable
  Int    i,j;
  Int64  iDist;
  Double dDist, sum;

  for(i=0; i< numCoef; i++)
  {
    dcoeff[i]= (Double)coef[i] / (Double)(1<< ((Int)AF_NUM_BIT_SHIFT) );
  }

  dDist =0;
  for(i=0; i< numCoef; i++)
  {
    sum= ((Double)eCorr[i][i]) * dcoeff[i];
    for(j=i+1; j< numCoef; j++)
    {
      sum += (Double)(2*eCorr[i][j])* dcoeff[j];
    }
    dDist += ((sum - 2.0 * yCorr[i])* dcoeff[i] );
  }

  UInt bitShift = g_uiBitIncrement<<1;
  if(dDist < 0)
  {
    iDist = -(((Int64)(-dDist + 0.5)) >> bitShift);
  }
  else //dDist >=0
  {
    iDist= ((Int64)(dDist+0.5)) >> bitShift;
  }

  return iDist;
}

Int TEncAdaptiveFilter::estimateFiltRate(AFFilter* filt, Int shape)
{
  Int rate= 1; //enable flag

  if(filt->enableFlag)
  {
    Int numCoef = g_AFNumCoef[shape];
    for(Int i=0; i< numCoef; i++)
    {
      rate += svlcBitrateEsitmate(filt->coeff[i]);
    }
  }

  return rate;
}

Void TEncAdaptiveFilter::predictCoef(Int* coef, Int shape /* =0 */)
{
  Int numCoef = g_AFNumCoef[shape];
  Int sum=0;
  
  for(Int i=0; i< numCoef-1;i++)
  {
    sum += (2* coef[i]);
  }

  Int pred = (1<<AF_NUM_BIT_SHIFT) - (sum);
  coef[numCoef-1] = coef[numCoef-1] - pred;
}

UInt TEncAdaptiveFilter::uvlcBitrateEstimate(Int val)
{
  val++;
  assert ( val );
  UInt length = 1;
  while( 1 != val )
  {
    val >>= 1;
    length += 2;
  }
  return ((length >> 1) + ((length+1) >> 1));
}

Void TEncAdaptiveFilter::quantFilterCoef(Double* h, Int* qh, Int* filtMag, Int N)
{
  Int i, maxValue, minValue;
  Double dblTotalGain;
  Int totalGain, qTotalGain;
  Int upper, lower;
  Double *dh;
  Int    *nc;

  dh = new Double[N];
  nc = new Int[N];
  maxValue =   (1<<(1+AF_NUM_BIT_SHIFT))-1;
  minValue = 0-(1<<(1+AF_NUM_BIT_SHIFT));
  dblTotalGain=0.0;
  qTotalGain=0;
  for(i=0; i<N; i++)
  {
    if(h[i]>=0.0)
    {
      qh[i] =  (Int)( h[i]*(1<<AF_NUM_BIT_SHIFT)+0.5);
    }
    else
    {
      qh[i] = -(Int)(-h[i]*(1<<AF_NUM_BIT_SHIFT)+0.5);
    }
    dh[i] = (Double)qh[i]/(Double)(1<<AF_NUM_BIT_SHIFT) - h[i];
    dh[i]*= filtMag[i];
    dblTotalGain += h[i]*filtMag[i];
    qTotalGain   += qh[i]*filtMag[i];
    nc[i] = i;
  }

  // modification of quantized filter coefficients
  totalGain = (Int)(dblTotalGain*(1<<AF_NUM_BIT_SHIFT)+0.5);
  if( qTotalGain != totalGain )
  {
    sortFilterCoef(dh, nc, 0, N-1);
    if( qTotalGain > totalGain )
    {
      upper = N-1;
      while( qTotalGain > totalGain+1 )
      {
        i = nc[upper%N];
        qh[i]--;
        qTotalGain -= filtMag[i];
        upper--;
      }
      if( qTotalGain == totalGain+1 )
      {
        if(dh[N-1]>0)
        {
          qh[N-1]--;
        }
        else
        {
          i=nc[upper%N];
          qh[i]--;
          qh[N-1]++;
        }
      }
    }
    else if( qTotalGain < totalGain )
    {
      lower = 0;
      while( qTotalGain < totalGain-1 )
      {
        i=nc[lower%N];
        qh[i]++;
        qTotalGain += filtMag[i];
        lower++;
      }
      if( qTotalGain == totalGain-1 )
      {
        if(dh[N-1]<0)
        {
          qh[N-1]++;
        }
        else
        {
          i=nc[lower%N];
          qh[i]++;
          qh[N-1]--;
        }
      }
    }
  }

  // set of filter coefficients
  for(i=0; i<N; i++)
  {
    qh[i] = max(minValue,min(maxValue, qh[i]));
  }

  checkFilterCoeffValue(qh, N);

  delete[] dh;
  dh = NULL;

  delete[] nc;
  nc = NULL;
}

Void TEncAdaptiveFilter::sortFilterCoef( Double *coefData, Int *coefNum, Int upper, Int lower )
{
  Double mid, tmpData;
  Int i, j, tmpNum;

  i = upper;
  j = lower;
  mid = coefData[(lower+upper)>>1];
  do
  {
    while( coefData[i] < mid ) i++;
    while( mid < coefData[j] ) j--;
    if( i <= j )
    {
      tmpData = coefData[i];
      tmpNum  = coefNum[i];
      coefData[i] = coefData[j];
      coefNum[i]  = coefNum[j];
      coefData[j] = tmpData;
      coefNum[j]  = tmpNum;
      i++;
      j--;
    }
  } while( i <= j );
  if ( upper < j ) 
  {
    sortFilterCoef(coefData, coefNum, upper, j);
  }
  if ( i < lower ) 
  {
    sortFilterCoef(coefData, coefNum, i, lower);
  }
}



//---------  Solve W-H equation ------------//
#define ROUND(a)  (((a) < 0)? (int)((a) - 0.5) : (int)((a) + 0.5))
#define REG              0.0001
#define REG_SQR          0.0000001

Int TEncAdaptiveFilter::gnsSolveByChol(Double **LHS, Double *rhs, Double *x, Int noEq)
{
  assert(noEq > 0);

  Double aux[AF_MAX_NUM_COEF];     /* Auxiliary vector */
  Double U[AF_MAX_NUM_COEF][AF_MAX_NUM_COEF];    /* Upper triangular Cholesky factor of LHS */
  Int  i, singular;          /* Looping variable */
  
  /* The equation to be solved is LHSx = rhs */
  
  /* Compute upper triangular U such that U'*U = LHS */
  if(gnsCholeskyDec(LHS, U, noEq)) /* If Cholesky decomposition has been successful */
  {
    singular = 1;
    /* Now, the equation is  U'*U*x = rhs, where U is upper triangular
     * Solve U'*aux = rhs for aux
     */
    gnsTransposeBacksubstitution(U, rhs, aux, noEq);         
    
    /* The equation is now U*x = aux, solve it for x (new motion coefficients) */
    gnsBacksubstitution(U, aux, noEq, x);   
    
  }
  else /* LHS was singular */ 
  {
    singular = 0;
    
    /* Regularize LHS */
    for(i=0; i < noEq; i++)
    {
      LHS[i][i] += REG;
    }
    /* Compute upper triangular U such that U'*U = regularized LHS */
    singular = gnsCholeskyDec(LHS, U, noEq);
    if ( singular == 1 )
    {
      /* Solve  U'*aux = rhs for aux */  
      gnsTransposeBacksubstitution(U, rhs, aux, noEq);   
      
      /* Solve U*x = aux for x */
      gnsBacksubstitution(U, aux, noEq, x);      
    }
    else
    {
      x[0] = 1.0;
      for (i = 1; i < noEq; i++ )
      {
        x[i] = 0.0;
      }
    }
  }  
  return singular;
}

Int TEncAdaptiveFilter::gnsCholeskyDec(Double **inpMatr, Double outMatr[AF_MAX_NUM_COEF][AF_MAX_NUM_COEF], Int noEq)
{ 
  Int i, j, k;     /* Looping Variables */
  Double scale;       /* scaling factor for each row */
  Double invDiag[AF_MAX_NUM_COEF];  /* Vector of the inverse of diagonal entries of outMatr */

  //  Cholesky decomposition starts

  for(i = 0; i < noEq; i++)
  {
    for(j = i; j < noEq; j++)
    {
      /* Compute the scaling factor */
      scale = inpMatr[i][j];
      if ( i > 0) 
      {
        for( k = i - 1 ; k >= 0 ; k--)
        {
          scale -= outMatr[k][j] * outMatr[k][i];
        }
      }
      /* Compute i'th row of outMatr */
      if(i == j)
      {
        if(scale <= REG_SQR ) // if(scale <= 0 )  /* If inpMatr is singular */
        {
          return 0;
        }
        else
        {
          /* Normal operation */
          invDiag[i] =  1.0 / (outMatr[i][i] = sqrt(scale));
        }
      }
      else
      {
        outMatr[i][j] = scale * invDiag[i]; /* Upper triangular part          */
        outMatr[j][i] = 0.0;              /* Lower triangular part set to 0 */
      }                    
    }
  }
  return 1; /* Signal that Cholesky factorization is successfully performed */
}


Void TEncAdaptiveFilter::gnsTransposeBacksubstitution(Double U[AF_MAX_NUM_COEF][AF_MAX_NUM_COEF], Double rhs[], Double x[], Int order)
{
  Int i,j;              /* Looping variables */
  Double sum;              /* Holds backsubstitution from already handled rows */

  /* Backsubstitution starts */
  x[0] = rhs[0] / U[0][0];               /* First row of U'                   */
  for (i = 1; i < order; i++)
  {         /* For the rows 1..order-1           */

    for (j = 0, sum = 0.0; j < i; j++) /* Backsubst already solved unknowns */
    {
      sum += x[j] * U[j][i];
    }
    x[i] = (rhs[i] - sum) / U[i][i];       /* i'th component of solution vect.  */
  }
}

Void  TEncAdaptiveFilter::gnsBacksubstitution(Double R[AF_MAX_NUM_COEF][AF_MAX_NUM_COEF], Double z[AF_MAX_NUM_COEF], Int R_size, Double A[AF_MAX_NUM_COEF])
{
  Int i, j;
  Double sum;

  R_size--;

  A[R_size] = z[R_size] / R[R_size][R_size];

  for (i = R_size-1; i >= 0; i--)
  {
    for (j = i + 1, sum = 0.0; j <= R_size; j++)
    {
      sum += R[i][j] * A[j];
    }

    A[i] = (z[i] - sum) / R[i][i];
  }
}
//---------  Solve W-H equation ------------//


#endif
//! \}
