/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TEncAdaptiveFilter.h
 \brief    estimation part of adaptive filter class (header)
 */
#ifndef __TENCADAPTIVEFILTER__
#define __TENCADAPTIVEFILTER__

#include "TLibCommon/TComAdaptiveFilter.h"

#if SVC_IL_ADAPTIVE_FILTER

//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// constant definition
// ====================================================================================================================


// ====================================================================================================================
// struct/enum
// ====================================================================================================================

struct AFCorrData
{
  Double** ECorr;
  Double*  yCorr;
  Double pixAcc;

  Void reset()
  {
    for(Int j=0; j< AF_MAX_NUM_COEF; j++)
    {
      for(Int i=0; i< AF_MAX_NUM_COEF; i++)
      {
        ECorr[j][i] = 0.0;
      }
      yCorr[j] = 0.0;
    }
    pixAcc = 0.0;
  }

  AFCorrData& operator += (const AFCorrData& src)
  {
    for(Int j=0; j< AF_MAX_NUM_COEF; j++)
    {
      for(Int i=0; i< AF_MAX_NUM_COEF; i++)
      {
        ECorr[j][i] += src.ECorr[j][i];
      }
      yCorr[j] += src.yCorr[j];
    }
    pixAcc += src.pixAcc;
    return *this;
  }
  AFCorrData()
  {
    ECorr = new Double*[AF_MAX_NUM_COEF];
    for(Int i=0; i< AF_MAX_NUM_COEF; i++)
    {
      ECorr[i]= new Double[AF_MAX_NUM_COEF];
    }
    yCorr = new Double[AF_MAX_NUM_COEF];
    reset();
  }
  ~AFCorrData()
  {
    for(Int i=0; i< AF_MAX_NUM_COEF; i++)
    {
      delete[] ECorr[i];
    }
    delete[] ECorr; ECorr = NULL;
    delete[] yCorr; yCorr = NULL;
  }
};


struct AFCTUInfo
{
  Int yPos;
  Int xPos;
  Int height;
  Int width;
};

// ====================================================================================================================
// class
// ====================================================================================================================
class TEncAdaptiveFilter: public TComAdaptiveFilter
{
public: 
  TEncAdaptiveFilter();
  ~TEncAdaptiveFilter();

  Void createAFEncData();  
  Void destroyAFEncData(); 
  Void predictCoef(Int* coef, Int shape =0); //need to called between AF and entropy coding stage

public: //inter-layer adaptive filtering: public interface  

#if SVC_ILAF_ONOFF_FLAG
  Void filterProcess(TComPicYuv* orgYuv, TComPicYuv* upYuv, TComPicYuv* filteredYuv, Double lambdaLuma, Double lambdaChroma);
#else
  Void filterProcess(TComPicYuv* orgYuv, TComPicYuv* upYuv, Double lambdaLuma, Double lambdaChroma);
#endif
#if SVC_ILAF_QD_MODEL
  Void estimateEarlyTermination(TComPicYuv* orgYuv, TComPicYuv* upYuv, Int baseQp, SliceType sliceType, Int enhanceQp,  Bool &earlyTermination);
#endif
  AFFilter&   getAFFilter     (Int compIdx)              {return m_afFilter[compIdx];}     // filters
  TComPicYuv* getAFFilteredPic()                         {return m_tempPicYuv;}            // filtered picture
private: //inter-layer adaptive filtering: private interface
  Void prepareEncData();
  Void getStatistics(AFCorrData* afCorr, TComPicYuv* orgYuv, TComPicYuv* srcYuv);
  Void estimateFilters(AFCorrData* afCorr, AFFilter* afFilter);
  Void filterPic(TComPicYuv* srcYuv, TComPicYuv* resYuv, AFFilter* afFilter);
  
private: //inter-layer adaptive filtering: member
  AFCorrData  m_afCorr   [AF_NUM_COMPONETS];    //correlation data for each CTB
  AFFilter    m_afFilter [AF_NUM_COMPONETS];
  Bool*       m_ctuEnableFlag;                  //CTU on/off flag
  AFCTUInfo*  m_ctuInfo;                        //CTU information: size, position

private: //common functions
  //statistics collection
  Void  getBlockCorr(AFCorrData* afCorr, Pel* orgPic, Pel* srcPic, Int stride, Int yPos, Int xPos, Int height, Int width, Bool isLastBlock, Int shape =0);

  //filter derivation
  Void  deriveFilter(AFCorrData* afCorr, Int* coef, Int shape= 0);
  Void  deriveOffFilter(Int* coef, Int shape=0);
  Void  quantFilterCoef(Double* h, Int* qh, Int* filtMag, Int N);
  Void  sortFilterCoef( Double *coef_data, Int *coef_num, Int upper, Int lower );

  //rate, distortion estimation
  UInt uvlcBitrateEstimate(Int val);
  UInt svlcBitrateEsitmate(Int val) {  return uvlcBitrateEstimate (( val <= 0) ? (-val<<1) : ((val<<1)-1));}
  Int64 estimateDistReduction(AFCorrData* afCorr, Int* coef, Int shape= 0);
  Int   estimateFiltRate(AFFilter* filt, Int shape = 0);

  //W-H. equation sol.
  Int   gnsSolveByChol(Double **LHS, Double *rhs, Double *x, Int noEq);
  Int   gnsCholeskyDec(Double **inpMatr, Double outMatr[AF_MAX_NUM_COEF][AF_MAX_NUM_COEF], Int noEq);
  Void  gnsTransposeBacksubstitution(Double U[AF_MAX_NUM_COEF][AF_MAX_NUM_COEF], Double rhs[], Double x[], Int order);
  Void  gnsBacksubstitution(Double R[AF_MAX_NUM_COEF][AF_MAX_NUM_COEF], Double z[AF_MAX_NUM_COEF], Int R_size, Double A[AF_MAX_NUM_COEF]);

private: //common variables
  Double   m_labmda[AF_NUM_COMPONETS];

};


#endif
#endif
